from typing import List

from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2 import Template as _Template

Template = _Template


def get_templates_environment(template_dirs: List[str]):
    return Environment(
        loader=FileSystemLoader(template_dirs),
        autoescape=select_autoescape(['html', 'xml']),
        enable_async=True,
    )



from typing import List, Callable, Coroutine

ASGIAppInstance = Callable[[Callable, Callable], Coroutine]
ASGIApp = Callable[[dict], ASGIAppInstance]
WSGIApp = Callable[[dict, Callable], List[bytes]]



import inspect
from http import HTTPStatus
from typing import Optional, List

from asgiref.sync import sync_to_async
from parse import parse

from bocadillo.exceptions import HTTPError
from .view import View, CallableView


class Route:
    """Represents a route to a view.

    Formatted string syntax is used for route patterns.
    """

    def __init__(self,
                 pattern: str,
                 view: View,
                 methods: List[str],
                 name: str):
        self._pattern = pattern
        self._view_is_class = inspect.isclass(view)
        if self._view_is_class:
            view = view()
        elif not inspect.iscoroutinefunction(view):
            view = sync_to_async(view)
        self._view = view
        self._methods = methods
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def url(self, **kwargs) -> str:
        """Return full path for the given route parameters."""
        return self._pattern.format(**kwargs)

    def match(self, path: str) -> Optional[dict]:
        """Return whether the route matches the given path.

        Examples
        -------
        >>> route = Route('/{age:d}', lambda req, resp: None)
        >>> route.match('/42')
        {'age': 42}
        >>> route.match('/john')
        None
        """
        result = parse(self._pattern, path)
        if result is not None:
            return result.named
        return None

    def _find_view(self, request) -> CallableView:
        """Find a view able to handle the request.

        Supports both function-based views and class-based views.
        """
        not_allowed_error = HTTPError(status=HTTPStatus.METHOD_NOT_ALLOWED)

        if self._view_is_class:
            if hasattr(self._view, 'handle'):
                view: CallableView = self._view.handle
            else:
                method_func_name = request.method.lower()
                view: CallableView = getattr(self._view, method_func_name, None)
                if view is None:
                    raise not_allowed_error
            if not inspect.iscoroutinefunction(view):
                view = sync_to_async(view)
        else:
            if request.method not in self._methods:
                raise not_allowed_error
            view = self._view

        return view

    async def __call__(self, request, response, **kwargs) -> None:
        view = self._find_view(request)
        await view(request, response, **kwargs)

    @property
    def name(self) -> str:
        return self._view.__name__



"""The Bocadillo API class."""
import os
from contextlib import contextmanager
from http import HTTPStatus
from typing import (Optional, Tuple, Type, List, Dict, Any, Union, Coroutine)

from asgiref.wsgi import WsgiToAsgi
from jinja2 import FileSystemLoader
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.testclient import TestClient
from uvicorn.main import run, get_logger
from uvicorn.reloaders.statreload import StatReload

from .checks import check_route
from .constants import ALL_HTTP_METHODS
from .cors import DEFAULT_CORS_CONFIG
from .error_handlers import ErrorHandler, handle_http_error
from .exceptions import HTTPError
from .middleware import CommonMiddleware, RoutingMiddleware
from .redirection import Redirection
from .request import Request
from .response import Response
from .route import Route
from .static import static
from .templates import Template, get_templates_environment
from .types import ASGIApp, WSGIApp, ASGIAppInstance


class API:
    """Bocadillo API.

    Parameters
    ----------
    templates_dir : str, optional
        The name of the directory containing templates, relative to
        the application entry point.
        Defaults to 'templates'.
    static_dir: str, optional
        The name of the directory containing static files, relative to
        the application entry point.
        Defaults to 'static'.
    static_root : str, optional
        The path prefix for static assets.
        Defaults to 'static'.
    allowed_hosts : list of str, optional
        A list of hosts which the server is allowed to run at.
        If the list contains '*', any host is allowed.
        Defaults to ['*'].
    enable_cors : bool, optional
        If True, Cross Origin Resource Sharing will be configured according
        to `cors_config`.
        Defaults to False.
    cors_config : dict, optional
        A dictionary of CORS configuration parameters.
        Defaults to `{'allow_origins': [], 'allow_methods': ['GET']}`.
        See also: https://www.starlette.io/middleware/#corsmiddleware
    enable_hsts : bool, optional
        If True, enable HSTS (HTTP Strict Transport Security) and automatically
        redirect HTTP traffic to HTTPS.
        Defaults to False.
    """

    _error_handlers: List[Tuple[Type[Exception], ErrorHandler]]

    def __init__(
            self,
            templates_dir: str = 'templates',
            static_dir: Optional[str] = 'static',
            static_root: Optional[str] = 'static',
            allowed_hosts: List[str] = None,
            enable_cors: bool = False,
            cors_config: dict = None,
            enable_hsts: bool = False,
    ):
        self._routes: Dict[str, Route] = {}
        self._named_routes: Dict[str, Route] = {}

        self._error_handlers = []
        self.add_error_handler(HTTPError, handle_http_error)

        self._templates = get_templates_environment([
            os.path.abspath(templates_dir),
        ])
        self._templates.globals.update(self._get_template_globals())

        self._extra_apps: Dict[str, Any] = {}

        self.client = self._build_client()

        if static_dir is not None:
            if static_root is None:
                static_root = static_dir
            self.mount(static_root, static(static_dir))

        if allowed_hosts is None:
            allowed_hosts = ['*']
        self.allowed_hosts = allowed_hosts

        if cors_config is None:
            cors_config = {}
        self.cors_config = {**DEFAULT_CORS_CONFIG, **cors_config}

        # Middleware
        self._routing_middleware = RoutingMiddleware(self)
        self._common_middleware = CommonMiddleware(self._routing_middleware)
        self.add_middleware(
            TrustedHostMiddleware,
            allowed_hosts=self.allowed_hosts,
        )
        if enable_cors:
            self.add_middleware(CORSMiddleware, **self.cors_config)
        if enable_hsts:
            self.add_middleware(HTTPSRedirectMiddleware)

    def _build_client(self) -> TestClient:
        return TestClient(self)

    def mount(self, prefix: str, app: Union[ASGIApp, WSGIApp]):
        """Mount another WSGI or ASGI app at the given prefix."""
        if not prefix.startswith('/'):
            prefix = '/' + prefix
        self._extra_apps[prefix] = app

    def add_error_handler(self, exception_cls: Type[Exception],
                          handler: ErrorHandler):
        """Register a new error handler.

        Parameters
        ----------
        exception_cls : Exception class
            The type of exception that should be handled.
        handler : (request, response, exception) -> None
            The actual error handler, which is called when an instance of
            `exception_cls` is caught.
        """
        self._error_handlers.insert(0, (exception_cls, handler))

    def error_handler(self, exception_cls: Type[Exception]):
        """Register a new error handler (decorator syntax).

        Example
        -------
        >>> import bocadillo
        >>> api = bocadillo.API()
        >>> @api.error_handler(KeyError)
        ... def on_key_error(req, resp, exc):
        ...     pass  # perhaps set resp.content and resp.status_code
        """

        def wrapper(handler):
            self.add_error_handler(exception_cls, handler)
            return handler

        return wrapper

    def _find_handlers(self, exception):
        return (
            handler for err_type, handler in self._error_handlers
            if isinstance(exception, err_type)
        )

    def _handle_exception(self, request, response, exception) -> None:
        """Handle an exception raised during dispatch.

        At most one handler is called for the exception: the first one
        to support it.

        If no handler was registered for the exception, it is raised.
        """
        for handler in self._find_handlers(exception):
            handler(request, response, exception)
            break
        else:
            raise exception from None

    def route(self, pattern: str, *, methods: List[str] = None,
              name: str = None):
        """Register a new route.

        Parameters
        ----------
        pattern : str
            A route pattern given as an f-string expression.
        methods : list of str, optional
            HTTP methods supported by this route.
            Defaults to all HTTP methods.
            Ignored for class-based views.
        name : str, optional
            A name for this route, which must be unique.

        Example
        -------
        >>> import bocadillo
        >>> api = bocadillo.API()
        >>> @api.route('/greet/{person}')
        ... def greet(req, resp, person: str):
        ...     pass
        """
        if methods is None:
            methods = ALL_HTTP_METHODS

        methods = [method.upper() for method in methods]

        def wrapper(view):
            check_route(pattern, view, methods)
            route = Route(
                pattern=pattern,
                view=view,
                methods=methods,
                name=name,
            )

            self._routes[pattern] = route
            if name is not None:
                self._named_routes[name] = route

            return route

        return wrapper

    def _find_matching_route(self, path: str) -> Tuple[Optional[str], dict]:
        """Find a route matching the given path."""
        for pattern, route in self._routes.items():
            kwargs = route.match(path)
            if kwargs is not None:
                return pattern, kwargs
        return None, {}

    def _get_route_or_404(self, name: str):
        try:
            return self._named_routes[name]
        except KeyError as e:
            raise HTTPError(HTTPStatus.NOT_FOUND.value) from e

    def url_for(self, name: str, **kwargs) -> str:
        """Return the URL path for a route.

        Parameters
        ----------
        name : str
            Name of the route.
        kwargs :
            Route parameters.

        Raises
        ------
        HTTPError(404) :
            If no route exists for the given `name`.
        """
        route = self._get_route_or_404(name)
        return route.url(**kwargs)

    def redirect(self, *,
                 name: str = None,
                 url: str = None,
                 permanent: bool = False,
                 **kwargs):
        """Redirect to another route.

        Parameters
        ----------
        name : str, optional
            Name of the route to redirect to.
        url : str, optional (unless name not given)
            URL of the route to redirect to.
        permanent : bool, optional
            If False (the default), returns a temporary redirection (302).
            If True, returns a permanent redirection (301).
        kwargs :
            Route parameters.
        """
        if name is not None:
            url = self.url_for(name=name, **kwargs)
        else:
            assert url is not None, 'url is expected if no route name is given'
        raise Redirection(url=url, permanent=permanent)

    def _get_template_globals(self) -> dict:
        return {
            'url_for': self.url_for,
        }

    @property
    def templates_dir(self) -> str:
        loader: FileSystemLoader = self._templates.loader
        return loader.searchpath[0]

    @templates_dir.setter
    def templates_dir(self, templates_dir: str):
        loader: FileSystemLoader = self._templates.loader
        loader.searchpath = [os.path.abspath(templates_dir)]

    def _get_template(self, name: str) -> Template:
        return self._templates.get_template(name)

    @contextmanager
    def _prevent_async_template_rendering(self):
        """If enabled, temporarily disable async template rendering.

        Notes
        -----
        Hot fix for a bug with Jinja2's async environment, which always
        renders asynchronously even under `render()`.
        Example error:
        `RuntimeError: There is no current event loop in thread [...]`
        """
        if not self._templates.is_async:
            yield
            return

        self._templates.is_async = False
        try:
            yield
        finally:
            self._templates.is_async = True

    @staticmethod
    def _prepare_context(context: dict = None, **kwargs):
        if context is None:
            context = {}
        context.update(kwargs)
        return context

    async def template(self, name_: str,
                       context: dict = None, **kwargs) -> Coroutine:
        """Render a template asynchronously.

        Can only be used within `async`  functions.

        Parameters
        ----------
        name_ : str
            Name of the template, located inside `templates_dir`.
            Trailing underscore to avoid collisions with a potential
            context variable named 'name'.
        context : dict
            Context variables to inject in the template.
        """
        context = self._prepare_context(context, **kwargs)
        return await self._get_template(name_).render_async(context)

    def template_sync(self, name_: str, context: dict = None, **kwargs) -> str:
        """Render a template synchronously.

        See Also
        --------
        .template()
        """
        context = self._prepare_context(context, **kwargs)
        with self._prevent_async_template_rendering():
            return self._get_template(name_).render(context)

    def template_string(self, source: str, context: dict = None,
                        **kwargs) -> str:
        """Render a template from a string (synchronous)."""
        context = self._prepare_context(context, **kwargs)
        with self._prevent_async_template_rendering():
            template = self._templates.from_string(source=source)
            return template.render(context)

    def run(self,
            host: str = None,
            port: int = None,
            debug: bool = False,
            log_level: str = 'info'):
        """Serve the application using uvicorn.

        Parameters
        ----------
        host : str, optional
            The host to bind to.
            Defaults to '127.0.0.1' (localhost). If not given and `PORT` is set,
            '0.0.0.0' will be used to serve to all known hosts.
        port : int, optional
            The port to bind to.
            Defaults to 8000 or (if set) the value of the `PORT` environment
            variable.
        debug : bool, optional
            Whether to serve the application in debug mode. Defaults to `False`.
        log_level : str, optional
            A logging level for the debug logger. Must be compatible with
            logging levels from the `logging` module.

        See Also
        --------
        https://www.uvicorn.org/settings/
        """
        if 'PORT' in os.environ:
            port = int(os.environ['PORT'])
            if host is None:
                host = '0.0.0.0'

        if host is None:
            host = '127.0.0.1'

        if port is None:
            port = 8000

        if debug:
            reloader = StatReload(get_logger(log_level))
            reloader.run(run, {
                'app': self,
                'host': host,
                'port': port,
                'log_level': log_level,
                'debug': debug,
            })
        else:
            run(self, host=host, port=port)

    async def dispatch(self, request: Request) -> Response:
        """Dispatch a request and return a response."""
        response = Response(request)

        try:
            pattern, kwargs = self._find_matching_route(request.url.path)
            route = self._routes.get(pattern)
            if route is None:
                raise HTTPError(status=404)
            else:
                try:
                    await route(request, response, **kwargs)
                except Redirection as redirection:
                    response = redirection.response
        except Exception as e:
            self._handle_exception(request, response, e)

        return response

    def _is_routing_middleware(self, middleware_cls) -> bool:
        return hasattr(middleware_cls, 'dispatch')

    def add_middleware(self, middleware_cls, **kwargs):
        if self._is_routing_middleware(middleware_cls):
            self._routing_middleware.add(middleware_cls, **kwargs)
        else:
            self._common_middleware.add(middleware_cls, **kwargs)

    def _find_app(self, scope: dict) -> ASGIAppInstance:
        """Return an ASGI app depending on the scope's path."""
        path: str = scope['path']

        # Return a sub-mounted extra app, if found
        for prefix, app in self._extra_apps.items():
            if not path.startswith(prefix):
                continue
            # Remove prefix from path so that the request is made according
            # to the mounted app's point of view.
            scope['path'] = path[len(prefix):]
            try:
                return app(scope)
            except TypeError:
                app = WsgiToAsgi(app)
                return app(scope)

        return self._common_middleware(scope)

    def __call__(self, scope: dict) -> ASGIAppInstance:
        return self._find_app(scope)



"""Bocadillo CLI factory."""
import os
from inspect import cleandoc
from typing import List

from .ext import click

CUSTOM_COMMANDS_FILE_ENV_VAR = 'BOCA_CUSTOM_COMMANDS_FILE'


def get_custom_commands_file_path():
    return os.getenv(CUSTOM_COMMANDS_FILE_ENV_VAR, 'boca.py')


class FileGroupCLI(click.MultiCommand):
    """Multi-command CLI that loads a group of commands from a file.

    Notes
    -----
    - Commands are loaded from a file determined by `file_name`, relative to the
    current working directory (i.e., when the CLI script is executed).
    - The first click.Group object found in that file is used as a source of
    commands.

    Inspired By
    -----------
    https://click.palletsprojects.com/en/7.x/commands/#custom-multi-commands
    """

    def __init__(self, file_name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._group: click.Group = None
        self.file_name = file_name

    def _load_group(self):
        ns = {}
        path = self.file_name

        try:
            with open(path, 'r') as f:
                code = compile(source=f.read(), filename=path, mode='exec')
                eval(code, ns, ns)
        except FileNotFoundError:
            # User did not create custom commands, use an empty group.
            self._group = click.Group()
            return

        for key, value in ns.items():
            if isinstance(value, click.Group):
                self._group = value
                break
        else:
            raise click.ClickException(
                f'Expected at least one group in {path}, none found.'
            )

    @property
    def group(self) -> click.Group:
        """Lazy-loaded, cached group object."""
        if self._group is None:
            self._load_group()
        return self._group

    def list_commands(self, ctx: click.Context) -> List[str]:
        return self.group.list_commands(ctx)

    def get_command(self, ctx: click.Context, name: str) -> click.Command:
        return self.group.get_command(ctx, name)


def create_cli() -> click.Command:
    @click.group()
    def builtin():
        pass

    @builtin.command(name='help')
    @click.pass_context
    def help_(ctx):
        """Show help about boca."""
        click.echo(ctx.parent.get_help())

    @builtin.command(name='init:custom')
    @click.option('-d', '--directory', default='',
                  help='Where files should be generated.')
    def init_custom(directory: str):
        """Generate files required to build custom commands."""
        custom_commands_script_contents = cleandoc(
            '''"""Custom Bocadillo commands.
    
            Use Click to build custom commands. For documentation, see:
            https://click.palletsprojects.com
            """
            from bocadillo.ext import click
    
    
            @click.group()
            def cli():
                pass
    
            # Write your @cli.command() functions below.\n
            '''
        ) + '\n'
        path = os.path.join(directory, get_custom_commands_file_path())
        with open(path, 'w') as f:
            f.write(custom_commands_script_contents)
        click.echo(click.style(f'Generated {path}', fg='green'))
        click.echo('Open the file and start building!')

    custom = FileGroupCLI(
        file_name=get_custom_commands_file_path(),
    )

    # Builtins first to prevent override in custom commands.
    return click.CommandCollection(sources=[builtin, custom])


# Exposed to the setup.py entry point
cli = create_cli()



from whitenoise import WhiteNoise

from .compat import empty_wsgi_app
from .types import WSGIApp


def static(directory: str) -> WSGIApp:
    """Return a WSGI app that serves static files under the given directory.

    Powered by WhiteNoise.
    """
    app = WhiteNoise(empty_wsgi_app())
    app.add_files(directory)
    return app



from http import HTTPStatus
from typing import Union
from jinja2.exceptions import TemplateNotFound as _TemplateNotFound

# Alias
TemplateNotFound = _TemplateNotFound


class RouteDeclarationError(Exception):
    """Raised when a route is ill-declared."""


class HTTPError(Exception):
    """Raised when an HTTP error occurs.

    You can raise this within a view or an error handler to interrupt
    request processing.
    """

    def __init__(self, status: Union[int, HTTPStatus]):
        if isinstance(status, int):
            status = HTTPStatus(status)
        else:
            assert isinstance(status, HTTPStatus), (
                f'Expected int or HTTPStatus, got {type(status)}'
            )
        self._status = status

    @property
    def http_status(self) -> HTTPStatus:
        return self._status

    @property
    def status_code(self) -> int:
        """Return the HTTP error's status code, i.e. 404."""
        return self._status.value

    @property
    def status_phrase(self) -> str:
        """Return the HTTP error's status phrase, i.e. 'Not Found'."""
        return self._status.phrase

    def __str__(self):
        return f'{self.status_code} {self.status_phrase}'


"""Built-in error handlers."""
from typing import Callable

from .exceptions import HTTPError
from .request import Request
from .response import Response


def handle_http_error(_, res, exc: HTTPError):
    res.status_code = exc.status_code
    res.html = f'<h1>{exc.status_code} {exc.status_phrase}</h1>'


ErrorHandler = Callable[[Request, Response, Exception], None]



import json
from typing import AnyStr

from starlette.requests import Request
from starlette.responses import Response as _Response


class Response:
    """Response builder."""

    def __init__(self, request: Request):
        self.request = request
        self._content: AnyStr = None
        self.status_code: int = None
        self.headers = {}

    def __setattr__(self, key, value):
        if key == 'text':
            self.headers['Content-Type'] = 'text/plain'
            self._content = value
        elif key == 'html':
            self.headers['Content-Type'] = 'text/html'
            self._content = value
        elif key == 'media':
            self.headers['Content-Type'] = 'application/json'
            self._content = json.dumps(value)
        elif key == 'content':
            self._content = value
        else:
            super().__setattr__(key, value)

    async def __call__(self, receive, send):
        """Build and send the response."""
        if self.status_code is None:
            self.status_code = 200

        if self.status_code != 204:
            self.headers.setdefault('Content-Type', 'text/plain')

        response = _Response(
            content=self._content,
            headers=self.headers,
            status_code=self.status_code,
        )
        await response(receive, send)



from typing import Callable, Type, Union, List, Optional

from .constants import ALL_HTTP_METHODS
from .request import Request
from .response import Response


class BaseView:
    """Base view interface."""

    def get(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def post(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def put(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def patch(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError

    def delete(self, request: Request, response: Response, **kwargs):
        raise NotImplementedError


CallableView = Callable[[Request, Response, dict], None]
ClassView = Type[BaseView]
View = Union[CallableView, ClassView]


def get_view_name(view: CallableView, base: Optional[ClassView] = None) -> str:
    return '.'.join(part.__name__ for part in (base, view) if part)


def get_declared_method_views(view_cls: ClassView) -> List[CallableView]:
    views: List[CallableView] = []
    for method in ALL_HTTP_METHODS:
        method_view_name = method.lower()
        if hasattr(view_cls, method_view_name):
            views.append(getattr(view_cls, method_view_name))
    return views



"""Bocadillo middleware definition.

A middleware is a class that implements the ASGI protocol.

There are two kinds of middleware classes:
- Routing middleware: suited for performing operations before and after
routing occurs.
- Common middleware: suited for all other middleware operations.

Common middleware should be called first, then routing middleware, which should
end up calling the actual Bocadillo API object.
"""
from .request import Request


class Middleware:
    """Base class for middleware classes."""

    def __init__(self, app):
        self.app = app

    def add(self, middleware_cls, **kwargs):
        self.app = middleware_cls(self.app, **kwargs)

    def __call__(self, scope: dict):
        raise NotImplementedError


class CommonMiddleware(Middleware):
    """Base middleware for non-routing operations."""

    def __call__(self, scope: dict):
        return self.app(scope)


class RoutingMiddleware(Middleware):
    """Helper for middleware that act before and after routing."""

    def before_dispatch(self, req):
        pass

    def after_dispatch(self, req, res):
        pass

    async def dispatch(self, request):
        self.before_dispatch(request)
        response = await self.app.dispatch(request)
        self.after_dispatch(request, response)
        return response

    def __call__(self, scope: dict):
        async def asgi(receive, send):
            nonlocal scope
            request = Request(scope, receive)
            response = await self.dispatch(request)
            await response(receive, send)

        return asgi



from .types import WSGIApp


def empty_wsgi_app() -> WSGIApp:
    """Return a WSGI app that always returns 404 Not Found."""

    def wsgi(environ, start_response):
        status = '404 Not Found'
        body = b'Not Found'
        headers = [('Content-Type', 'text/plain')]
        start_response(status, headers)
        return [body]

    return wsgi



"""Package setup."""

import setuptools

description = (
    'A modern Python web framework filled with asynchronous salsa.'
)

with open('README.md', 'r') as readme:
    long_description = readme.read()


setuptools.setup(
    name='bocadillo',
    version='0.5.0',
    author='Florimond Manca',
    author_email='florimond.manca@gmail.com',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['bocadillo', 'bocadillo.ext'],
    install_requires=[
        'starlette',
        'uvicorn',
        'jinja2',
        'asgiref',
        'whitenoise',
        'requests',
        'parse',
    ],
    url='https://github.com/bocadilloproject/bocadillo',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    entry_points={
        'console_scripts': [
            'boca=bocadillo.cli:cli',
        ]
    },
)



"""Start-up check and validation utilities."""
import inspect
from string import Formatter
from typing import List

from .constants import ALL_HTTP_METHODS
from .exceptions import RouteDeclarationError
from .view import get_declared_method_views, View, get_view_name


def check_route(pattern: str, view: View, methods: List[str]) -> None:
    """Check compatibility of a route pattern and a view."""
    _check_methods(view, methods)
    _check_route_pattern(pattern, view)
    _check_route_parameters(pattern, view)


def _check_methods(view: View, methods: List[str]) -> None:
    for method in methods:
        if method not in ALL_HTTP_METHODS:
            raise RouteDeclarationError(
                f'Route "{view.__name__}" accepts method "{method}" '
                'but it is not one of the valid HTTP methods: '
                f'{", ".join(ALL_HTTP_METHODS)}'
            )


def _check_route_pattern(pattern: str, view: View) -> None:
    if not pattern.startswith('/'):
        raise RouteDeclarationError(
            f'Route pattern "{pattern}" on view "{view.__name__}" '
            f'must start with "/" to avoid ambiguities.'
        )


def _check_route_parameters(pattern: str, view: View, _base=None) -> None:
    """Verify that a view accepts parameters defined in a route pattern.

    Raises
    ------
    RouteDeclarationError :
        If a route parameter is not one of the view's arguments,
        of a view argument was not declared in the route pattern.
    """

    parsed_format = Formatter().parse(pattern)
    route_parameters: set = {name for _, name, _, _ in parsed_format if name}

    if inspect.isclass(view):
        views = get_declared_method_views(view)
        for method_view in views:
            _check_route_parameters(pattern, method_view, _base=view)
    else:
        view_name = get_view_name(view=view, base=_base)
        view_parameters = dict(inspect.signature(view).parameters)

        # Necessary if view is from a class-based view.
        view_parameters.pop('self', None)

        for route_param in route_parameters:
            if route_param not in view_parameters:
                raise RouteDeclarationError(
                    f'Parameter "{route_param}" was declared on route '
                    f'"{view_name}()" and should be one of '
                    'its arguments.'
                )

        if len(view_parameters) < 2:
            raise RouteDeclarationError(
                f'View "{view_name}" must have at least two '
                'parameters (request and response).'
            )

        for i, view_param in enumerate(view_parameters):
            is_req_or_res = i < 2
            if is_req_or_res:
                continue
            if view_param not in route_parameters:
                raise RouteDeclarationError(
                    f'Parameter "{view_param}" is expected by route '
                    f'"{view_name}" but was not declared in the '
                    'route pattern.'
                )
